"""
:synopsis: The main npu api
.. moduleauthor:: Naval Bhandari <naval@neuro-ai.co.uk>
"""

from .npu import api, predict, compile, train, export
from . import vision, optim, loss

api.__module__ = "npu"
predict.__module__ = "npu"
compile.__module__ = "npu"
train.__module__ = "npu"
export.__module__ = "npu"



import hashlib
import requests

from .saving.saving import save_data
from .web.urls import TOKEN_URL, HASH_URL, UPLOAD_DATA_URL
from .Dataset import Dataset

_token = ""


def api(token_):
    global _token
    _token = token_
    params = {"token": _token}
    response = requests.get(TOKEN_URL, params=params)
    if response.status_code == 200:
        print("Token successfully authenticated")
        return response
    else:
        raise ValueError(response.text)
    # "API token not valid"


def getToken():
    return _token


def checkData(data):
    if isinstance(data, (str, dict)):
        return data
    elif isinstance(data, Dataset):
        return data
    else:
        return uploadData(data).json()


def sliceData(data):
    id = data["id"]
    x = data["indexes"]
    y = None
    if isinstance(x, slice):
        y = x.stop
        x = x.start
    return id, x, y


def uploadData(data):
    print("Saving data locally...")
    if isinstance(data, str):
        file = open(data, "rb")
    else:
        file = save_data(data)
    print("Hashing...")
    hash = hashlib.md5(file.read()).hexdigest()
    print("Checking if data is on servers...")
    params = {"token": getToken(), "hash": hash, "collection": 1}
    response = requests.get(HASH_URL, params=params)
    if response.status_code == 200:
        file.close()
        return response
    print("Data not found on servers. Uploading now...")
    file.seek(0)
    files = {'file': file}
    response = requests.post(UPLOAD_DATA_URL, files=files, params=params)
    if isinstance(data, str):
        file.close()
    return response

def hashData(data):
    pass












import glob
import tarfile
import os
from tempfile import TemporaryFile
import numpy as np


def save_model(model, library: str):
    model_path = model
    if not isinstance(model, str):
        if not os.path.isdir("tmp"):
            os.mkdir("tmp")
        model_path = "tmp/tmp_model"
        if library == "pytorch":
            import torch
            import dill
            model_path += ".pt"
            torch.save(model, model_path, pickle_module=dill)
        elif library == "keras":
            model_path += ".h5"
            import keras
            with keras.backend.get_session().graph.as_default():
                model.save_model(model_path)
        elif library == "mxnet":
            model_path += ".tar"
            model.export(model_path)
            # with tarfile.open(model_path, "w") as tar:
            #     for file in glob.glob(model_path + "dir/*"):
            #         tar.add(file, arcname=os.path.basename(file))
            with tarfile.open(model_path, "w") as tar:
                jsonname = model_path + "-symbol.json"
                paramname = model_path + "-0000.params"
                tar.add(jsonname, arcname=os.path.basename(jsonname))
                tar.add(paramname, arcname=os.path.basename(paramname))
        elif library == "TF":
            model_path += ".tar"
            model.save(model_path + "dir")
            with tarfile.open(model_path, "w") as tar:
                for file in glob.glob(model_path + "dir/*"):
                    tar.add(file, arcname=os.path.basename(file))
        elif library == "TF1":
            model_path += ".pb"
        else:
            raise ValueError("Model type: " + str(library) + " not defined")
    return model_path

    # if model_type is ModelType.ONNX:
    #     onnx.save(model, file_path)
    # elif model_type is ModelType.TF1:
    #     pass
        # tf.saved_model.save(model, "./dd.pb")
        # tf.compat.v1.saved_model.save(model, "tmp")
        # raise ValueError("Tensorflow 1 incompatible. Please use .pb file directly if using Tensorflow 1.")


def save_data(data):
    file = TemporaryFile()
    if isinstance(data, (tuple, list)):
        if not all([isinstance(x, np.ndarray) for x in data]):
            raise TypeError("Either train or validation data is not in numpy array format.")
        dict_data = {str(i): data[i] for i in range(0, len(data))}
        np.savez_compressed(file, **dict_data)
    else:
        if not isinstance(data, np.ndarray):
            raise TypeError("Data is not in numpy array format.")
        np.savez_compressed(file, data)
    file.seek(0)
    return file


def determine_model(model):
    try:
        from torch import nn
        if isinstance(model, nn.Module):
            return "pytorch"
    finally:
        pass
    try:
        from tensorflow.keras import Model
        if isinstance(model, Model):
            return "TF"
    finally:
        pass
    try:
        import mxnet
        if isinstance(model, mxnet.gluon.nn.Block):
            return "mxnet"
    finally:
        return ""



"""
:synopsis: Loss functions to use with NPU API for training
.. moduleauthor:: Naval Bhandari <naval@neuro-ai.co.uk>
"""

CrossEntropyLoss = "CrossEntropyLoss"  #:
SparseCrossEntropyLoss = "SparseCrossEntropyLoss"  #:
MSELoss = "MSELoss"  #:
CTCLoss = "CTCLoss"  #:
L1Loss = "L1Loss"  #:
NLLLoss = "NLLLoss"  #:



import hashlib

import bson
import requests

from .common import getToken
from .Task import Task
from .saving.saving import determine_model
from .web.urls import RETRIEVE_MODEL_URL, HASH_URL, COMPILE_URL
from .saving import save_model


def compile(model, input_shape, library, model_label, asynchronous=False):
    if model_label != "" and model is None:
        print(model_label)
        params = {"token": getToken(), "input_shape": input_shape, "label": model_label}
        response = requests.get(RETRIEVE_MODEL_URL, params=params)
        if response.status_code == 200:
            return response
        else:
            raise LookupError("Model not found. " + str(response))
    if bson.ObjectId.is_valid(model):
        return model
    print("Saving model locally.")
    if library == "":
        library = determine_model(model)
    file_path = save_model(model, library)
    print("Model saved locally.")
    with open(file_path, "rb") as file:
        hash = hashlib.md5(file.read()).hexdigest()
        if model_label == "":
            model_label = file_path
        params = {"token": getToken(), "input_shape": input_shape, "given_name": model_label, "hash": hash,
                  "collection": 2, "modelType": library}
        response = requests.get(HASH_URL, params=params)
        if response.status_code == 200:
            return response.json()
        elif response.status_code != 204:
            raise ConnectionAbortedError("Checking hash not worked. {0}".format(response.content))
        else:
            print("Model not found on server. Uploading now...")
        task = Task(compileApi(file, params))
        if not asynchronous:
            task.wait()
            print("Model compiled successfully")
        return task


def compileApi(file, params):
    file.seek(0)
    files = {'file': file}
    response = requests.post(COMPILE_URL, files=files, params=params)
    return response.json()




"""
:synopsis: Global datasets available
.. moduleauthor:: Naval Bhandari <naval@neuro-ai.co.uk>
"""

from ..core import Dataset

CIFAR10 = Dataset("cifar10_train", 40000, 50000, 60000)
""":CIFAR10: CIFAR10 dataset. Contains training/validation/test subsets of data. Accessed using 
`CIFAR10.train/val/test` properties. Can also get direct subset using slices. """



"""
:synopsis: Global models available
.. moduleauthor:: Naval Bhandari <naval@neuro-ai.co.uk>
"""

from ..core import Model

resnet18 = Model("resnet18")
""":resnet18: Resnet 18 model. Access pretrained version `resnet18(pretrained=True)` """

vgg16 = Model("vgg16")
""":vgg16: VGG16 model. Access pretrained version `vgg16(pretrained=True)` """





# @wait_task
import requests

from .Dataset import Dataset
from .common import getToken, checkData, sliceData
from .Task import Task
from .web.urls import TRAIN_URL
from .Model import Model


def train(model, training_data, test_data, batch_size, epochs, optimiser, loss, asynchronous, callback):
    training_data = checkData(training_data)
    test_data = checkData(test_data)
    task_id = model.task_id if isinstance(model, Task) else ""
    task = Task(trainApi(model, training_data, test_data, optimiser, loss, batch_size, epochs, task_id).json(),
                callback)
    if not asynchronous:
        task.wait(show_progress=True)
        print("Model finished training")
    return task


def trainApi(model_id, train_id, test_id, optimiser, loss, batch_size, epochs, task_id=""):
    train_x = train_y = test_x = test_y = None
    train_name = ""
    test_name = ""
    if isinstance(train_id, dict):
        train_id, train_x, train_y = sliceData(train_id)
    if isinstance(test_id, dict):
        test_id, test_x, test_y = sliceData(test_id)
    if isinstance(train_id, Dataset):
        train_name = train_id.id
        train_id = ""
    if isinstance(test_id, Dataset):
        test_name = test_id.id
        test_id = ""
    params = {"trainId": train_id, "testId": test_id, "loss": loss,
              "token": getToken(), "batch_size": batch_size, "epochs": epochs, "task_id": task_id, "train_x": train_x,
              "train_y": train_y, "test_x": test_x, "test_y": test_y, "train_name": train_name, "test_name": test_name}

    if isinstance(model_id, Model):
        params["model_name"] = model_id.name
    elif model_id != "" and not isinstance(model_id, Task):
        params["modelId"] = model_id
    response = requests.get(TRAIN_URL, params=params, json=optimiser)
    if response.status_code != 200:
        raise ValueError(response.text)
    return response



# @wait_task
import requests

from .common import getToken, checkData, sliceData
from .Task import Task
from .web.urls import PREDICT_URL
from .Dataset import Dataset
from .Model import Model


def predict(model, data, asynchronous=False, callback=None):
    inference_data = checkData(data)
    task_id = model.task_id if isinstance(model, Task) else ""
    resp = predictApi(model, inference_data, task_id)
    task = Task(resp.json(), callback)
    if not asynchronous:
        task.wait()
    return task


def predictApi(model, data_id, task_id):
    x = y = None
    data_name = ""
    if isinstance(data_id, dict):
        data_id, x, y = sliceData(data_id)
    if isinstance(data_id, Dataset):
        data_name = data_id.id
        data_id = ""
    params = {"dataId": data_id, "token": getToken(), "task_id": task_id, "x": x, "y": y, "data_name": data_name}
    if isinstance(model, Model):
        params["model_name"] = model.name
    elif model != "" and not isinstance(model, Task):
        params["modelId"] = model
    response = requests.get(PREDICT_URL, params=params)
    if response.status_code != 200:
        raise ValueError(response.text)
    return response


